library AC;

{ Network Drawing program }

{ By Phil Marty <phmarty@yahoo.fr> }

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

{
This is then DC engine plugin to use whith Fendi
It is the first very simple DC simulation engine as exemple
}

uses
  SysUtils,
  Classes;

const
   MaxLines=1000;

   {$I '../CompoConst.lst'}

type
  TLigInteger=array[0..999999] of Integer;
  PLigInteger=^TLigInteger;

  TLigDouble=array[0..999999] of Double;
  PLigDouble=^TLigDouble;
  TMatDouble=array[0..999999] of PLigDouble;
  PMatDouble=^TMatDouble;

  // Type used to communication between Fendi and the engines
  // It is quite a text file
  TComm=record
    Count:Integer; // 32 bits
    Lines:array[0..MaxLines-1] of Pchar;
    end;
  PComm=^TComm; // Pointer

  // Common father
  TObjetSimu=class
  end;

  // The net nodes
  TNodeSimu=class(TObjetSimu)
    Number:Integer;
    Father:TObjetSimu;
    procedure SetNumber(_Number:Integer);
    function  GetNumber:Integer;
    procedure SetFather(_Father:TObjetSimu);
    function  GetFather:TObjetSimu;
  end;

  // The components
  TComponentSimu=class(TObjetSimu)
    Name:string;
    ObjetType:Integer;
    NodesList:TList;
    NbParameter:Integer;
    ParameterName:array[0..10] of string; // Will be dynamic later
    Parameter:array[0..10] of Double;     // Will be dynamic later
    PosMatrix:Integer;

    constructor Create;
    destructor Destroy;

    procedure SetName(NewName:string);
    function  GetName:string;
    procedure SetType(NewType:Integer);
    function  GetType:Integer;

    procedure SetNode(NodeSimu:TNodeSimu;NodeIndex:Integer);
    function  GetNode(i:Integer):TNodeSimu;
    function  GetNodeCount:Integer;

    procedure SetParametersCount(Count:Integer);
    function  GetParametersCount:Integer;
    procedure SetParameterName(i:Integer;_ParameterName:string);
    function  GetParameterName(i:Integer):string;
    procedure SetParameterValue(i:Integer;Value:Double);
    function  GetParameterValue(i:Integer):Double;

    procedure SetPosMatrix(Position:Integer);
    function  GetPosMatrix:Integer;
  end;

var
   SaveExit: Pointer;

   NetNodesList:TList;
   ListOfComponentsSimu:TList;
   NbActiveComponents:Integer;
   ListActive:TList;
   NbNodes:Integer;
   PositionNodes:Integer;   
   ErrorMsg:string;

   MatDouble1:PMatDouble;
   MatDouble2:PMatDouble;

//   MatDoubleReal1:PMatDouble;
//   MatDoubleReal2:PMatDouble;
//   MatDoubleImag1:PMatDouble;
//   MatDoubleImag2:PMatDouble;

   Frequency:Double;

{$R *.RES}

// Allocation matrix
procedure GetmemMat(var MatDouble:PMatDouble;
                    Sx,Sy:Integer);
var
   j:Integer;
begin
Getmem(MatDouble,Sy*4);
for j:=0 to Sy-1 do
   begin
   Getmem(MatDouble^[j],Sx*8);
   FillChar(MatDouble^[j]^,Sx*8,0);
   end;
end;

// Free matrix
procedure FreememMat(var MatDouble:PMatDouble;
                     Sx,Sy:Integer);
var
   j:Integer;
begin
for j:=0 to Sy-1 do Freemem(MatDouble^[j],Sx*8);
Freemem(MatDouble,Sy*4);
end;

function GaussJordan(var a:PMatDouble; n:Integer; var b:PMatDouble; m:Integer):Boolean;
var
big,dum,pivinv:Double;
i,icol,irow,j,k,l,ll:Integer;
indxc,indxr,ipiv:PLigInteger;
begin
Getmem(indxc,n*8);
Getmem(indxr,n*8);
Getmem(ipiv,n*8);
Result:=True;
for j:=0 to n-1 do ipiv^[j]:=0;
for i:=0 to n-1 do
   begin
   big:=0;
   for j:=0 to n-1 do
      if ipiv[j]<>1 then
         for k:=0 to n-1 do
            if ipiv^[k]=0 then
               if abs(a^[j]^[k])>=big then
                  begin
                  big:=abs(a^[j]^[k]);
                  irow:=j;
                  icol:=k;
                  end
               else if ipiv^[k]>1 then begin Result:=False; Exit; end;
   ipiv^[icol]:=ipiv^[icol]+1;
   if irow<>icol then
      begin
      for l:=0 to n-1 do
         begin
         dum:=a^[irow]^[l];
         a^[irow]^[l]:=a^[icol]^[l];
         a^[icol]^[l]:=dum;
         end;
      for l:=0 to m-1 do
         begin
         dum:=b^[irow]^[l];
         b^[irow]^[l]:=b^[icol]^[l];
         b^[icol]^[l]:=dum;
         end;
      end;
   indxr^[i]:=irow;
   indxc^[i]:=icol;
   if abs(a[icol,icol])<1e-18 then begin Result:=False; Exit; end;
   pivinv:=1/a^[icol]^[icol];
   a^[icol]^[icol]:=1;
   for l:=0 to n-1 do a^[icol]^[l]:=a^[icol]^[l]*pivinv;
   for l:=0 to m-1 do b^[icol]^[l]:=b^[icol]^[l]*pivinv;
   for ll:=0 to n-1 do
      if ll<>icol then
         begin
         dum:=a^[ll]^[icol];
         a^[ll]^[icol]:=0;
         for l:=0 to n-1 do
            a^[ll]^[l]:=a^[ll]^[l]-a^[icol]^[l]*dum;
         for l:=0 to m-1 do
            b^[ll]^[l]:=b^[ll]^[l]-b^[icol]^[l]*dum;
         end;
   end;
for l:=n-1 downto 0 do
   if indxr^[l]<>indxc^[l] then
      for k:=0 to n-1 do
         begin
         dum:=a^[k]^[indxr^[l]];
         a^[k]^[indxr^[l]]:=a^[k]^[indxc^[l]];
         a^[k]^[indxc^[l]]:=dum;
         end;
Result:=True;
end;

//******************************************************************************
//***************************    TNodeSimu    **********************************
//******************************************************************************

procedure TNodeSimu.SetNumber(_Number:Integer);
begin
Number:=_Number;
end;

function TNodeSimu.GetNumber:Integer;
begin
Result:=Number;
end;

procedure TNodeSimu.SetFather(_Father:TObjetSimu);
begin
Father:=_Father;
end;

function TNodeSimu.GetFather:TObjetSimu;
begin
Result:=Father;
end;

//******************************************************************************
//***********************       TComponentSimu     *****************************
//******************************************************************************

constructor TComponentSimu.Create;
begin
NodesList:=TList.Create;
end;

destructor TComponentSimu.Destroy;
begin
NodesList.Free;
end;

function TComponentSimu.GetName:string;
begin
Result:=Name;
end;

procedure TComponentSimu.SetName(NewName:string);
begin
Name:=NewName;
end;

procedure TComponentSimu.SetType(NewType:Integer);
begin
ObjetType:=NewType;
end;

function TComponentSimu.GetType:Integer;
begin
Result:=ObjetType;
end;

procedure TComponentSimu.SetNode(NodeSimu:TNodeSimu;NodeIndex:Integer);
begin
while NodeIndex>NodesList.Count-1 do NodesList.Add(nil);
NodesList[NodeIndex]:=NodeSimu;
end;

function TComponentSimu.GetNode(i:Integer):TNodeSimu;
begin
Result:=NodesList[i];
end;

function TComponentSimu.GetNodeCount:Integer;
begin
Result:=NodesList.Count;
end;

procedure TComponentSimu.SetParametersCount(Count:Integer);
begin
NbParameter:=Count;
end;

function TComponentSimu.GetParametersCount:Integer;
begin
Result:=NbParameter;
end;

procedure TComponentSimu.SetParameterName(i:Integer;_ParameterName:string);
begin
ParameterName[i]:=_ParameterName;
end;

function TComponentSimu.GetParameterName(i:Integer):string;
begin
Result:=ParameterName[i];
end;

procedure TComponentSimu.SetParameterValue(i:Integer;Value:Double);
begin
Parameter[i]:=Value;
end;

function TComponentSimu.GetParameterValue(i:Integer):Double;
begin
Result:=Parameter[i];
end;

procedure TComponentSimu.SetPosMatrix(Position:Integer);
begin
PosMatrix:=Position;
end;

function TComponentSimu.GetPosMatrix:Integer;
begin
Result:=PosMatrix;
end;

//******************************************************************************
//**************************  DC Simulation Engine   ***************************
//******************************************************************************

// Load data in Comm
//function Load(var Comm:PComm):Boolean;
function Load:Boolean;
var
   T:TextFile;
   i,j,k:Integer;
   PosX,PosY:integer;
   Path,Line:string;
   NetNode:TList;
   Typ:Integer;
   CompoName:string;
   ComponentSimu:TComponentSimu;
   NodeList:TList;
   NodeIndex:Integer;
   Father:TComponentSimu;
   NodeSimu:TNodeSimu;
begin
AssignFile(T,'datas.tmp');
Reset(T);
try

//k:=0;
Result:=True;

//<?xml version='1.0' encoding='UTF16' standalone='yes'?>
//Inc(k);
Readln(T,Line);

//<component>
//Line:=Comm.Lines[k];
//Inc(k);
Readln(T,Line);
if Line<>'<component>' then
   begin
   ErrorMsg:='Error : No component in the file';
   Result:=False;
   Exit;
   end;

// Load components
while Line='<component>' do
   begin
   //<type>Objet.GetType</type>
   //Line:=Comm.Lines[k];
   //Inc(k);
   Readln(T,Line);
   Typ:=StrToInt(Copy(Line,7,Length(Line)-13));
   //<name>Objet.GetName</name>
   //Line:=Comm.Lines[k];
   //Inc(k);
   Readln(T,Line);
   CompoName:=Copy(Line,7,Length(Line)-13);

   ComponentSimu:=TComponentSimu.Create;
   ComponentSimu.SetType(Typ);
   ComponentSimu.SetName(CompoName);
   ListOfComponentsSimu.Add(ComponentSimu);

   //<parameternumber>Objet.GetParametersCount</parameternumber>
   //Line:=Comm.Lines[k];
   //Inc(k);
   Readln(T,Line);
   Line:=Copy(Line,18,Length(Line)-35);
   ComponentSimu.SetParametersCount(StrToInt(Line));
   for i:=0 to ComponentSimu.GetParametersCount-1 do
      begin
      //<parametername>Objet.GetParameterName()</parametername>
      //Line:=Comm.Lines[k];
      //Inc(k);
      Readln(T,Line);
      Line:=Copy(Line,16,Length(Line)-31);
      ComponentSimu.SetParameterName(i,Line);
      //<parametervalue>Objet.GetParameter()</parametervalue>
      //Line:=Comm.Lines[k];
      //Inc(k);
      Readln(T,Line);
      Line:=Copy(Line,17,Length(Line)-33);
      ComponentSimu.SetParameterValue(i,StrToFloat(Line));
      end;

   //Inc(k);
   Readln(T,Line);
   //Line:=Comm.Lines[k];
   //Inc(k);
   Readln(T,Line);
   end;

// Load nodes and links to components
//<netnode>
if Line<>'<netnode>' then
   begin
   ErrorMsg:='Error : No netnode in the file';
   Result:=False;
   Exit;
   end;

while Line='<netnode>' do
   begin
   //<node>
   //Line:=Comm.Lines[k];
   //Inc(k);
   Readln(T,Line);
   if Line<>'<node>' then
      begin
      ErrorMsg:='Error : No node in netnode';
      Result:=False;
      Exit;
      end;

   NodeList:=TList.Create;
   NetNodesList.Add(NodeList);
   while Line='<node>' do
      begin
      //<component>Component.GetName</component>
      //Line:=Comm.Lines[k];
      //Inc(k);
      Readln(T,Line);
      CompoName:=Copy(Line,12,Length(Line)-23);
      //<index>Node.GetIndex</index>
      //Line:=Comm.Lines[k];
      //Inc(k);
      Readln(T,Line);
      NodeIndex:=StrToInt(Copy(Line,8,Length(Line)-15));

      // Get ComponentSimu with name CompoName
      for i:=0 to ListOfComponentsSimu.Count-1 do
         begin
         ComponentSimu:=ListOfComponentsSimu[i];
         if ComponentSimu.GetName=CompoName then Father:=ComponentSimu;
         end;

      NodeSimu:=TNodeSimu.Create;
      NodeSimu.SetNumber(NodeIndex);
      NodeSimu.SetFather(Father);
      Father.SetNode(NodeSimu,NodeIndex);
      NodeList.Add(NodeSimu);

      //Inc(k);
      Readln(T,Line);
      //Line:=Comm.Lines[k];
      //Inc(k);
      Readln(T,Line);
      end;
   //if k<Comm.Count then Line:=Comm.Lines[k];
   //Inc(k);
   Readln(T,Line);
   end;

finally
CloseFile(T);
end;
end;

// Save results in Comm
//function SaveResults(var Comm:PComm):Boolean;
function SaveResults:Boolean;
var
   i,Position:Integer;
   ComponentSimu:TComponentSimu;
   NetNode:TList;
   T:TextFile;

   procedure AddLine(StrVal:string);
   begin
//   Getmem(Comm.Lines[Comm.Count],Length(StrVal));
//   StrCopy(Comm.Lines[Comm.Count],PChar(StrVal));
//   Inc(Comm.Count);
   Writeln(T,StrVal);
   end;

begin
AssignFile(T,'results.tmp');
Rewrite(T);
try


//New(Comm);
//Comm.Count:=0;

AddLine('<?xml version=''1.0'' encoding=''UTF16'' standalone=''yes''?>');

// Save components infos
for i:=0 to NbActiveComponents-1 do
   begin
   ComponentSimu:=ListActive[i];
   Position:=ComponentSimu.GetPosMatrix;
   AddLine('<component>');
   AddLine('<type>'+IntToStr(ComponentSimu.GetType)+'</type>');
   AddLine('<name>'+ComponentSimu.GetName+'</name>');
   case ComponentSimu.GetType of
      ncTransformer:begin
         AddLine('<infosnumber>2</infosnumber>');

         AddLine('<infoname>Current1</infoname>');
         AddLine('<infotype>1</infotype>'); // Complex
         AddLine('<infounit>A</infounit>');
         AddLine('<infovalue>'+FloatToStr(MatDouble2^[Position  ]^[0])+'</infovalue>');
         AddLine('<infovalue>'+FloatToStr(MatDouble2^[Position+1]^[0])+'</infovalue>');

         AddLine('<infoname>Current3</infoname>');
         AddLine('<infotype>1</infotype>'); // Complex
         AddLine('<infounit>A</infounit>');
         AddLine('<infovalue>'+FloatToStr(MatDouble2^[Position+4]^[0])+'</infovalue>');
         AddLine('<infovalue>'+FloatToStr(MatDouble2^[Position+5]^[0])+'</infovalue>');
         end;
      ncZLine:begin
         AddLine('<infosnumber>3</infosnumber>');

         AddLine('<infoname>Current1</infoname>');
         AddLine('<infotype>1</infotype>'); // Complex
         AddLine('<infounit>A</infounit>');
         AddLine('<infovalue>'+FloatToStr(MatDouble2^[Position  ]^[0])+'</infovalue>');
         AddLine('<infovalue>'+FloatToStr(MatDouble2^[Position+1]^[0])+'</infovalue>');

         AddLine('<infoname>Current2</infoname>');
         AddLine('<infotype>1</infotype>'); // Complex
         AddLine('<infounit>A</infounit>');
         AddLine('<infovalue>'+FloatToStr(MatDouble2^[Position+2]^[0])+'</infovalue>');
         AddLine('<infovalue>'+FloatToStr(MatDouble2^[Position+3]^[0])+'</infovalue>');

         AddLine('<infoname>Current3</infoname>');
         AddLine('<infotype>1</infotype>'); // Complex
         AddLine('<infounit>A</infounit>');
         AddLine('<infovalue>'+FloatToStr(MatDouble2^[Position+4]^[0])+'</infovalue>');
         AddLine('<infovalue>'+FloatToStr(MatDouble2^[Position+5]^[0])+'</infovalue>');
         end
      else
         begin
         AddLine('<infosnumber>1</infosnumber>');
         AddLine('<infoname>Current</infoname>');
         AddLine('<infotype>1</infotype>'); // Complex
         AddLine('<infounit>A</infounit>');
         AddLine('<infovalue>'+FloatToStr(MatDouble2^[Position  ]^[0])+'</infovalue>');
         AddLine('<infovalue>'+FloatToStr(MatDouble2^[Position+1]^[0])+'</infovalue>');
         end;
      end;
   AddLine('</component>');
   end;

// Save nodes infos
for i:=0 to NbNodes-1 do
   begin
   NetNode:=NetNodesList[i];
   // Only first node of the net node / The others are the same
   AddLine('<netnode>');
   AddLine('<infosnumber>1</infosnumber>');
   AddLine('<infoname>Voltage</infoname>');
   AddLine('<infotype>1</infotype>'); // Complex
   AddLine('<infounit>V</infounit>');
//   AddLine('<infovalue>'+FloatToStr(MatDouble2^[(NbActiveComponents*4+i*2)]^[0])+'</infovalue>');
//   AddLine('<infovalue>'+FloatToStr(MatDouble2^[(NbActiveComponents*4+i*2)+1]^[0])+'</infovalue>');
   AddLine('<infovalue>'+FloatToStr(MatDouble2^[PositionNodes+i*2]^[0])+'</infovalue>');
   AddLine('<infovalue>'+FloatToStr(MatDouble2^[PositionNodes+i*2+1]^[0])+'</infovalue>');
   AddLine('</netnode>');
   end;

// Todo : Lines ?

finally
CloseFile(T);
end;
end;

//function Execute(var Comm:PComm):Boolean;
function Execute:Boolean;
var
   i,Position,PosEquation,j,k,l:Integer;
   ComponentSimu:TComponentSimu;
   Dim:Integer;
   Node1,Node2,Node3,Node4,Node5,Node6:TNodeSimu;
   NetNode:TList;
   NodeNumber1,NodeNumber2,NodeNumber3,NodeNumber4,NodeNumber5,NodeNumber6:Integer;
   CompoNumber,NeutralNode,CompoType:Integer;
   T:TextFile;
   LL,MM:Double;
begin
// Free old data
for i:=0 to ListOfComponentsSimu.Count-1 do
   begin
   ComponentSimu:=ListOfComponentsSimu[i];
   ComponentSimu.Free;
   end;
ListOfComponentsSimu.Clear;

for j:=0 to NetNodesList.Count-1 do
   begin
   NetNode:=NetNodesList[j];
   NetNode.Free;
   end;
NetNodesList.Clear;

// Load new data
//if not Load(Comm) then begin Result:=False; Exit; end;
if not Load then begin Result:=False; Exit; end;

// Free Comm
//for i:=0 to Comm.Count-1 do
//   StrDispose(Comm.Lines[i]);
//Comm.Count:=0;

// More than one neutral forbidden
j:=0;
for i:=0 to ListOfComponentsSimu.Count-1 do
   begin
   ComponentSimu:=ListOfComponentsSimu[i] ;
   if ComponentSimu.GetType=ncNeutral then Inc(j);
   end;

if j>1 then
   begin
   Result:=False;
   ErrorMsg:='Error : Only one neutral allowed';
   Exit;
   end;

// List of active components
ListActive:=TList.Create;

// Find active components and their position in matrix
j:=0;
Position:=0;
for i:=0 to ListOfComponentsSimu.Count-1 do
   begin
   ComponentSimu:=ListOfComponentsSimu[i];
   if ComponentSimu.GetType<>ncNeutral then
      begin
      Inc(j);
      ComponentSimu.SetPosMatrix(Position);
      Position:=Position+ComponentSimu.GetNodeCount*2;
//      case ComponentSimu.GetType of
//         ncTransformer:Position:=Position+8; //i1a1 i1b1 i1a2 i1b2 i1a3 i1b3 i1a4 i1b4
//         else
//            Position:=Position+4;  //i1a1 i1b1 i1a2 i1b2
//         end;
      ListActive.Add(ComponentSimu);
      end;
   end;

// Save position of first node
PositionNodes:=Position;
// Save number of nodes
NbNodes:=NetNodesList.Count;
// Save number of active components
NbActiveComponents:=j;
// Find matrix dimension
//Dim:=(NbNodes+NbActiveComponents*2)*2;
Dim:=Position+NbNodes*2;

// Create network matrix in the heap
GetmemMat(MatDouble1,Dim,Dim);
GetmemMat(MatDouble2,1,Dim);

// Matrix form :
// i1a1 i1b1 i1a2 i1b2 i2a1 i2b1 i2a2 i2b2 i3a1 i3b1 i3a2 i3b2......  v1a1 v1b1 v2a1 v2b1 v3a1 v3b1 ......
// where :
//                         |----------|
//           ^----->-------|          |-------<-----^
//  vax+j*vbx|  iax+j*ibx  |----------|   iay+j*iby |vay+jvby
//           |                                      |

// Resistor :
//                         |----------|
//           ^----->-------|    R     |-------<-----^
//  va1+j*vb1|  ia1+j*ib1  |----------|   ia2+j*ib2 |va2+jvb2
//           |                                      |
//
// R(ia1+j*ib1)=(va1+j*vb1)-(va2+jvb2)
//
// 4 Equations :
// R*ia1=va1-va2
// R*ib1=vb1-vb2
// ia1+ia2=0
// ib1+ib2=0

// Inductor :
//                         |----------|
//           ^----->-------|    Lw    |-------<-----^
//  va1+j*vb1|  ia1+j*ib1  |----------|   ia2+j*ib2 |va2+jvb2
//           |                                      |
//
// jLw(ia1+j*ib1)=(va1+j*vb1)-(va2+jvb2)
//
// 4 Equations :
// -Lw*ib1=va1-va2
//  Lw*ia1=vb1-vb2
// ia1+ia2=0
// ib1+ib2=0

// Capacitor :
//                         |----------|
//           ^----->-------|   1/Cw   |-------<-----^
//  va1+j*vb1|  ia1+j*ib1  |----------|   ia2+j*ib2 |va2+jvb2
//           |                                      |
//
// 1/jCw*(ia1+j*ib1)=(va1+j*vb1)-(va2+jvb2)
//
// 4 Equations :
//  1/Cw*ib1=va1-va2
// -1/Cw*ia1=vb1-vb2
// ia1+ia2=0
// ib1+ib2=0

// Transformer :
//                        |----------|
//          ^----->-------|          |-----<-----^
// va1+j*vb1|  ia1+j*ib1  |          | ia3+j*ib3 |va3+jvb3
//          |             |    k     |           |
//                        |          |
//          ^----->-------|          |-----<-----^
// va2+j*vb2|  ia2+j*ib2  |----------| ia4+j*ib4 |va4+jvb4
//          |                                    |
//
// (va3+jvb3)-(va4+jvb4)=k((va1+j*vb1)-(va2+j*vb2))
//  ia1+j*ib1=k(ia3+j*ib3)
//
// 8 Equations :
// k*va1-k*va2-va3+va4=0
// k*vb1-k*vb2-vb3+vb4=0
// ia1-k*ia3=0
// ib1-k*ib3=0
// ia1+ia2=0
// ib2+ib2=0
// ia3+ia4=0
// ib3+ib4=0

// Line :
//                        |----------|
//          ^----->-------|          |-----<-----^
// va1+j*vb1|  ia1+j*ib1  |          | ia4+j*ib4 |va4+jvb4
//          |             |          |           |
//                        |          |
//          ^----->-------|   Line   |-----<-----^
// va2+j*vb2|  ia2+j*ib2  |          | ia5+j*ib5 |va4+jvb5
//          |             |          |           |
//                        |          |
//          ^----->-------|          |-----<-----^
// va3+j*vb3|  ia3+j*ib3  |----------| ia6+j*ib6 |va6+jvb6
//          |                                    |
//

// (va1+jvb1)-(va4+jvb4)=jwL*(ia1+j*ib1)+jwM*((ia2+j*ib2)+(ia3+j*ib3))
// va1+jvb1-va4-jvb4    =jwL*ia1-wL*ib1+jwM*ia2-wM*ib2+jwM*ia3-wM*ib3
// (va1-va4)+j(vb1-vb4) =-wL*ib1-wM*ib2-wM*ib3+jwL*ia1+jwM*ia2+jwM*ia3
// (va1-va4)+j(vb1-vb4)=(-wL*ib1-wM*ib2-wM*ib3)+j(wL*ia1+wM*ia2+wM*ia3)

// (va2+jvb2)-(va5+jvb5)=jwL*(ia2+j*ib2)+jwM*((ia1+j*ib1)+(ia3+j*ib3))
// va2+jvb2-va5-jvb5)   =jwL*ia3-wL*ib3+jwM*ia1-wM*ib1+jwM*ia2-wM*ib2
// (va2-va5)+j(vb2-vb5) =-wL*ib3-wM*ib1-wM*ib2+jwL*ia3+jwM*ia1+jwM*ia2
// (va2-va5)+j(vb2-vb5) =(-wL*ib3-wM*ib1-wM*ib2)+j(wL*ia3+wM*ia1+wM*ia2)

// (va3+jvb3)-(va6+jvb6)=jwL*(ia3+j*ib3)+jwM*((ia1+j*ib1)+(ia2+j*ib2))
// va3+jvb3-va6-jvb6=jwL*ia3-wL*ib3+jwM*ia1-wM*ib1++jwM*ia2-wM*ib2
// (va3-va6)+j(vb3-vb6)=-wL*ib3-wM*ib1-wM*ib2+jwL*ia3+jwM*ia1+jwM*ia2
// (va3-va6)+j(vb3-vb6)=(-wL*ib3-wM*ib1-wM*ib2)+j(wL*ia3+wM*ia1+wM*ia2)

//
// 12 Equations :
// va1-va4+wL*ib1+wM*ib2+wM*ib3=0
// vb1-vb4-wL*ia1-wM*ia2-wM*ia3=0
// va2-va5+wL*ib3+wM*ib1+wM*ib2=0
// vb2-vb5-wL*ia3-wM*ia1-wM*ia2=0
// va3-va6+wL*ib3+wM*ib1+wM*ib2=0
// vb3-vb6-wL*ia3-wM*ia1-wM*ia2=0
// ia1+ia4=0
// ib1+ib4=0
// ia2+ia5=0
// ib2+ib5=0
// ia3+ia6=0
// ib3+ib6=0



//*****************************************
//********** Component equations **********
//*****************************************
// Position of current equation = matrix line
PosEquation:=0;
for i:=0 to NbActiveComponents-1 do
   begin
   ComponentSimu:=TComponentSimu(ListActive[i]);
   Position:=ComponentSimu.GetPosMatrix;

   // Find nodes
   Node1:=ComponentSimu.GetNode(0);
   Node2:=ComponentSimu.GetNode(1);
   if ComponentSimu.GetNodeCount>2 then Node3:=ComponentSimu.GetNode(2);
   if ComponentSimu.GetNodeCount>3 then Node4:=ComponentSimu.GetNode(3);
   if ComponentSimu.GetNodeCount>4 then Node5:=ComponentSimu.GetNode(4);
   if ComponentSimu.GetNodeCount>5 then Node6:=ComponentSimu.GetNode(5);
   // Find corresponding netnodes
   NodeNumber1:=-1;
   NodeNumber2:=-1;
   NodeNumber3:=-1;
   NodeNumber4:=-1;
   NodeNumber5:=-1;
   NodeNumber6:=-1;
   for j:=0 to NetNodesList.Count-1 do
      begin
      NetNode:=NetNodesList[j];
      for k:=0 to NetNode.Count-1 do
         begin
         if Node1=NetNode[k] then NodeNumber1:=j;
         if Node2=NetNode[k] then NodeNumber2:=j;
         if ComponentSimu.GetNodeCount>2 then
            if Node3=NetNode[k] then NodeNumber3:=j;
         if ComponentSimu.GetNodeCount>3 then
            if Node4=NetNode[k] then NodeNumber4:=j;
         if ComponentSimu.GetNodeCount>4 then
            if Node5=NetNode[k] then NodeNumber5:=j;
         if ComponentSimu.GetNodeCount>5 then
            if Node6=NetNode[k] then NodeNumber6:=j;
         end;
      end;

   case ComponentSimu.GetType of
      ncResistor:begin
         if (NodeNumber1=-1) or (NodeNumber2=-1) then
            begin
            ErrorMsg:='Error : I don''t find resistor nodes';
            Result:=False;
            Exit;
            end;
         // R*ia1-va1+va2=0
         MatDouble1^[PosEquation]^[Position                   ]:= ComponentSimu.GetParameterValue(0);
         MatDouble1^[PosEquation]^[PositionNodes+NodeNumber1*2]:=-1;
         MatDouble1^[PosEquation]^[PositionNodes+NodeNumber2*2]:= 1;

         // R*ib1-vb1+vb2=0
         MatDouble1^[PosEquation+1]^[Position+1                   ]:= ComponentSimu.GetParameterValue(0);
         MatDouble1^[PosEquation+1]^[PositionNodes+NodeNumber1*2+1]:=-1;
         MatDouble1^[PosEquation+1]^[PositionNodes+NodeNumber2*2+1]:= 1;

         PosEquation:=PosEquation+2;
         end;
      ncInductor:begin
         if (NodeNumber1=-1) or (NodeNumber2=-1) then
            begin
            ErrorMsg:='Error : I don''t find inductor nodes';
            Result:=False;
            Exit;
            end;
         // -Lw*ib1-va1+va2=0
         MatDouble1^[PosEquation]^[Position+1                 ]:=-ComponentSimu.GetParameterValue(0)*Frequency*2*pi;
         MatDouble1^[PosEquation]^[PositionNodes+NodeNumber1*2]:=-1;
         MatDouble1^[PosEquation]^[PositionNodes+NodeNumber2*2]:= 1;

         //  Lw*ia1-vb1+vb2=0
         MatDouble1^[PosEquation+1]^[Position                     ]:= ComponentSimu.GetParameterValue(0)*Frequency*2*pi;
         MatDouble1^[PosEquation+1]^[PositionNodes+NodeNumber1*2+1]:=-1;
         MatDouble1^[PosEquation+1]^[PositionNodes+NodeNumber2*2+1]:= 1;

         PosEquation:=PosEquation+2;
         end;
      ncCapacitor:begin
         if (NodeNumber1=-1) or (NodeNumber2=-1) then
            begin
            ErrorMsg:='Error : I don''t find capacitor nodes';
            Result:=False;
            Exit;
            end;
         //  1/Cw*ib1-va1+va2=0
         MatDouble1^[PosEquation]^[Position+1                 ]:= 1/ComponentSimu.GetParameterValue(0)/Frequency/2/pi;
         MatDouble1^[PosEquation]^[PositionNodes+NodeNumber1*2]:=-1;
         MatDouble1^[PosEquation]^[PositionNodes+NodeNumber2*2]:= 1;

         // -1/Cw*ia1-vb1+vb2=0
         MatDouble1^[PosEquation+1]^[Position]  :=-1/ComponentSimu.GetParameterValue(0)/Frequency/2/pi;
         MatDouble1^[PosEquation+1]^[PositionNodes+NodeNumber1*2+1]:=-1;
         MatDouble1^[PosEquation+1]^[PositionNodes+NodeNumber2*2+1]:= 1;

         PosEquation:=PosEquation+2;
         end;
      ncTransformer:begin
         if (NodeNumber1=-1) or (NodeNumber2=-1) or (NodeNumber3=-1) or (NodeNumber4=-1) then
            begin
            ErrorMsg:='Error : I don''t find transformer nodes';
            Result:=False;
            Exit;
            end;
         // ia1-k*ia3=0
         MatDouble1^[PosEquation]  ^[Position]  := 1;
         MatDouble1^[PosEquation]  ^[Position+4]:=-ComponentSimu.GetParameterValue(0);
         // ib1-k*ib3=0
         MatDouble1^[PosEquation+1]^[Position+1]:= 1;
         MatDouble1^[PosEquation+1]^[Position+5]:=-ComponentSimu.GetParameterValue(0);

         // k*va1-k*va2-va3+va4=0
         MatDouble1^[PosEquation+2]^[PositionNodes+NodeNumber1*2]  := ComponentSimu.GetParameterValue(0);
         MatDouble1^[PosEquation+2]^[PositionNodes+NodeNumber2*2]  :=-ComponentSimu.GetParameterValue(0);
         MatDouble1^[PosEquation+2]^[PositionNodes+NodeNumber3*2]  :=-1;
         MatDouble1^[PosEquation+2]^[PositionNodes+NodeNumber4*2]  := 1;
         // k*vb1-k*vb2-vb3+vb4=0
         MatDouble1^[PosEquation+3]^[PositionNodes+NodeNumber1*2+1]:= ComponentSimu.GetParameterValue(0);
         MatDouble1^[PosEquation+3]^[PositionNodes+NodeNumber2*2+1]:=-ComponentSimu.GetParameterValue(0);
         MatDouble1^[PosEquation+3]^[PositionNodes+NodeNumber3*2+1]:=-1;
         MatDouble1^[PosEquation+3]^[PositionNodes+NodeNumber4*2+1]:= 1;

         PosEquation:=PosEquation+4;
         end;
      ncZLine:begin
         if (NodeNumber1=-1) or (NodeNumber2=-1) or (NodeNumber3=-1) or (NodeNumber4=-1)
             or (NodeNumber5=-1) or (NodeNumber6=-1) then
            begin
            ErrorMsg:='Error : I don''t find line nodes';
            Result:=False;
            Exit;
            end;
         LL:=(2*ComponentSimu.GetParameterValue(0)+ComponentSimu.GetParameterValue(1))/3;
         MM:=(ComponentSimu.GetParameterValue(1)-ComponentSimu.GetParameterValue(0))/3;

         // va1-va4+wL*ib1+wM*ib2+wM*ib3=0
         MatDouble1^[PosEquation]^[PositionNodes+NodeNumber1*2]:= 1;
         MatDouble1^[PosEquation]^[PositionNodes+NodeNumber4*2]:=-1;
         MatDouble1^[PosEquation]^[Position+1]:=LL;
         MatDouble1^[PosEquation]^[Position+3]:=MM;
         MatDouble1^[PosEquation]^[Position+5]:=MM;
         // vb1-vb4-wL*ia1-wM*ia2-wM*ia3=0
         MatDouble1^[PosEquation+1]^[PositionNodes+NodeNumber1*2+1]:= 1;
         MatDouble1^[PosEquation+1]^[PositionNodes+NodeNumber4*2+1]:=-1;
         MatDouble1^[PosEquation+1]^[Position  ]:=-LL;
         MatDouble1^[PosEquation+1]^[Position+2]:=-MM;
         MatDouble1^[PosEquation+1]^[Position+3]:=-MM;
         // va2-va5+wL*ib3+wM*ib1+wM*ib2=0
         MatDouble1^[PosEquation+2]^[PositionNodes+NodeNumber2*2]:= 1;
         MatDouble1^[PosEquation+2]^[PositionNodes+NodeNumber5*2]:=-1;
         MatDouble1^[PosEquation+2]^[Position+1]:=MM;
         MatDouble1^[PosEquation+2]^[Position+3]:=LL;
         MatDouble1^[PosEquation+2]^[Position+5]:=MM;
         // vb2-vb5-wL*ia3-wM*ia1-wM*ia2=0
         MatDouble1^[PosEquation+3]^[PositionNodes+NodeNumber2*2+1]:= 1;
         MatDouble1^[PosEquation+3]^[PositionNodes+NodeNumber5*2+1]:=-1;
         MatDouble1^[PosEquation+3]^[Position  ]:=-MM;
         MatDouble1^[PosEquation+3]^[Position+2]:=-LL;
         MatDouble1^[PosEquation+3]^[Position+4]:=-MM;
         // va3-va6+wL*ib3+wM*ib1+wM*ib2=0
         MatDouble1^[PosEquation+4]^[PositionNodes+NodeNumber3*2]:= 1;
         MatDouble1^[PosEquation+4]^[PositionNodes+NodeNumber6*2]:=-1;
         MatDouble1^[PosEquation+4]^[Position+1]:=MM;
         MatDouble1^[PosEquation+4]^[Position+3]:=MM;
         MatDouble1^[PosEquation+4]^[Position+5]:=LL;
         // vb3-vb6-wL*ia3-wM*ia1-wM*ia2=0
         MatDouble1^[PosEquation+5]^[PositionNodes+NodeNumber3*2+1]:= 1;
         MatDouble1^[PosEquation+5]^[PositionNodes+NodeNumber6*2+1]:=-1;
         MatDouble1^[PosEquation+5]^[Position  ]:=-MM;
         MatDouble1^[PosEquation+5]^[Position+2]:=-MM;
         MatDouble1^[PosEquation+5]^[Position+4]:=-LL;

         PosEquation:=PosEquation+6;
         end;
      ncSourceAC:begin
         MatDouble2^[PosEquation]  ^[0]:=ComponentSimu.GetParameterValue(0)*
            Cos(ComponentSimu.GetParameterValue(1)*Pi/180);
         MatDouble1^[PosEquation]  ^[PositionNodes+NodeNumber1*2]  := 1;
         MatDouble1^[PosEquation]  ^[PositionNodes+NodeNumber2*2]  :=-1;

         MatDouble2^[PosEquation+1]^[0]:=ComponentSimu.GetParameterValue(0)*
            Sin(ComponentSimu.GetParameterValue(1)*Pi/180);
         MatDouble1^[PosEquation+1]^[PositionNodes+NodeNumber1*2+1]:= 1;
         MatDouble1^[PosEquation+1]^[PositionNodes+NodeNumber2*2+1]:=-1;

         PosEquation:=PosEquation+2;
         end;
      end;
   if PosEquation>Dim-1 then
      begin
      ErrorMsg:='Error : Equation position exceeds matrix dimension';
      Result:=False;
      Exit;
      end;
   end;

//**************************************************
//********** Components current equations **********
//**************************************************

// ia1 ->0
// ib1 ->1
// ia2 ->2
// ib2 ->3
// ia3 ->4
// ib3 ->5
// ia4 ->6
// ib4 ->7
// ia5 ->8
// ib5 ->9
// ia6 ->10
// ib6 ->11

for i:=0 to NbActiveComponents-1 do
   begin
   ComponentSimu:=TComponentSimu(ListActive[i]);
   Position:=ComponentSimu.GetPosMatrix;
   // For this firts DC engine, we take the first parameter of components
   // because there is only one parameter in components.
   case ComponentSimu.GetType of
      ncTransformer:begin
         // ia1+ia2=0
         MatDouble1^[PosEquation]  ^[Position]  :=1;
         MatDouble1^[PosEquation]  ^[Position+2]:=1;
         // ib1+ib2=0
         MatDouble1^[PosEquation+1]^[Position+1]:=1;
         MatDouble1^[PosEquation+1]^[Position+3]:=1;
         // ia3+ia4=0
         MatDouble1^[PosEquation+2]^[Position+4]:=1;
         MatDouble1^[PosEquation+2]^[Position+6]:=1;
         // ib3+ib4=0
         MatDouble1^[PosEquation+3]^[Position+5]:=1;
         MatDouble1^[PosEquation+3]^[Position+7]:=1;
         PosEquation:=PosEquation+4;
         if PosEquation>Dim-1 then
            begin
            ErrorMsg:='Error : Equation position exceeds matrix dimension';
            Result:=False;
            Exit;
            end;
         end;
      ncZline:begin
         // ia1+ia4=0
         MatDouble1^[PosEquation]^[Position]  :=1;
         MatDouble1^[PosEquation]^[Position+6]:=1;
         // ib1+ib4=0
         MatDouble1^[PosEquation+1]^[Position+1]:=1;
         MatDouble1^[PosEquation+1]^[Position+7]:=1;
         // ia2+ia5=0
         MatDouble1^[PosEquation+2]^[Position+2]:=1;
         MatDouble1^[PosEquation+2]^[Position+8]:=1;
         // ib2+ib5=0
         MatDouble1^[PosEquation+3]^[Position+3]:=1;
         MatDouble1^[PosEquation+3]^[Position+9]:=1;
         // ia3+ia6=0
         MatDouble1^[PosEquation+4]^[Position+4]  :=1;
         MatDouble1^[PosEquation+4]^[Position+10]:=1;
         // ib3+ib6=0
         MatDouble1^[PosEquation+5]^[Position+5]  :=1;
         MatDouble1^[PosEquation+5]^[Position+11]:=1;

         PosEquation:=PosEquation+6;
         if PosEquation>Dim-1 then
            begin
            ErrorMsg:='Error : Equation position exceeds matrix dimension';
            Result:=False;
            Exit;
            end;
         end
      else
         begin
         // ia1+ia2=0
         MatDouble1^[PosEquation]  ^[Position]  :=1;
         MatDouble1^[PosEquation]  ^[Position+2]:=1;
         // ib1+ib2=0         
         MatDouble1^[PosEquation+1]^[Position+1]:=1;
         MatDouble1^[PosEquation+1]^[Position+3]:=1;

         PosEquation:=PosEquation+2;
         if PosEquation>Dim-1 then
            begin
            ErrorMsg:='Error : Equation position exceeds matrix dimension';
            Result:=False;
            Exit;
            end;
         end;
      end;
   end;

//************************************
//********** Node equations **********
//************************************
for i:=0 to NbNodes-2 do
   begin
   // Find components connected to the node
   NetNode:=NetNodesList[i];
   for j:=0 to NetNode.Count-1 do
      begin
      Node1:=NetNode[j];
      ComponentSimu:=Node1.Father as TComponentSimu;
      if (ComponentSimu.GetType<>ncConnection) and (ComponentSimu.GetType<>ncNeutral) then
         begin
         // Find component number
         l:=0;
         for k:=0 to NbActiveComponents-1 do
            begin
            ComponentSimu:=TComponentSimu(ListActive[k]);
            if ComponentSimu=Node1.Father then
               begin
//               CompoNumber:=l;
               CompoNumber:=ComponentSimu.GetPosMatrix;
               CompoType:=ComponentSimu.GetType;
               end;
            inc(l);
            end;

         case CompoType of
            ncSourceAC:
               begin
               MatDouble1^[PosEquation]  ^[CompoNumber+Node1.GetNumber*2]  :=-1;
               MatDouble1^[PosEquation+1]^[CompoNumber+Node1.GetNumber*2+1]:=-1;
               end;
            ncTransformer:
               begin
               MatDouble1^[PosEquation]  ^[CompoNumber+Node1.GetNumber*2]  :=1;
               MatDouble1^[PosEquation+1]^[CompoNumber+Node1.GetNumber*2+1]:=1;
               end
            else
               begin
               MatDouble1^[PosEquation]  ^[CompoNumber+Node1.GetNumber*2]  :=1;
               MatDouble1^[PosEquation+1]^[CompoNumber+Node1.GetNumber*2+1]:=1;
               end
            end;
         end;
      end;
   PosEquation:=PosEquation+2;
   if PosEquation>Dim-1 then
      begin
      ErrorMsg:='Error : Equation position exceeds matrix dimension';
      Result:=False;
      Exit;
      end;
   end;

//**************************************
//********** Neutral equation **********
//**************************************
// Last special line for the neutral node
// Find neutral node
NeutralNode:=-1;
for i:=0 to NbNodes-1 do
   begin
   // Find components connected to the node
   NetNode:=NetNodesList[i];
   for j:=0 to NetNode.Count-1 do
      begin
      Node1:=NetNode[j];
      // Save neutral node for later
      ComponentSimu:=Node1.Father as TComponentSimu;
      if ComponentSimu.GetType=ncNeutral then
         NeutralNode:=i;
      end;
   end;

if NeutralNode=-1 then
   begin
   ErrorMsg:='Error : I don''t find ground node';
   Result:=False;
   Exit;
   end;

MatDouble1^[PosEquation]  ^[PositionNodes+NeutralNode*2]:=1;
MatDouble1^[PosEquation+1]^[PositionNodes+NeutralNode*2+1]:=1;

// Save matrix to a file to check
AssignFile(T,'Matrix.txt');
Rewrite(T);
try

for i:=0 to Dim-1 do
   begin
   for j:=0 to Dim-1 do
      Write(T,Format('%2.1g',[MatDouble1^[i]^[j]])+' ');
   Write(T,Format('%2.1g',[MatDouble2^[i]^[0]])+' ');
   Writeln(T,'');
   end;

// Solve network matrix
GaussJordan(MatDouble1,Dim,MatDouble2,1);

Writeln(T,'');
for i:=0 to Dim-1 do
   begin
   for j:=0 to Dim-1 do
      Write(T,Format('%2.1g',[MatDouble1^[i]^[j]])+' ');
   Write(T,Format('%2.1g',[MatDouble2^[i]^[0]])+' ');
   Writeln(T,'');
   end;

finally
CloseFile(T);
end;

// Save results
//SaveResults(Comm);
SaveResults;

ListActive.Free;
end;

// List of allowed components for this engine
function AllowedComponent(Number:Integer):Boolean;
begin
Result:=False;
case Number of
   ncNone       :Result:=True;
   ncNode       :Result:=True;
   ncLine       :Result:=True;
   ncConnection :Result:=True;
   ncCapacitor  :Result:=True;
   ncResistor   :Result:=True;
   ncInductor   :Result:=True;
   ncSourceAC   :Result:=True;
   ncNeutral    :Result:=True;
   ncComment    :Result:=True;
   ncTransformer:Result:=True;
   ncZLine      :Result:=True;
   ncMultipole  :Result:=True;      
   end;
end;

function GetParamCount:Integer; 
begin
Result:=1;
end;

function GetParamName(i:Integer):PChar;
begin
if i=0 then Result:='Frequency' else Result:='';
end;

function GetParamUnit(i:Integer):PChar;
begin
if i=0 then Result:='Hz' else Result:='';
end;

procedure SetParamValue(i:Integer;Value:Double);
begin
if i=0 then Frequency:=Value;
end;

function GetParamDefaultValue(i:Integer):Double;
begin
Result:=50;
end;

function GetErrorMsg:PChar;
begin
Result:=PChar(ErrorMsg);
end;

procedure LibExit;
begin
// Library exit code
ListOfComponentsSimu.Free;
NetNodesList.Free;
ExitProc := SaveExit;  // Restore output procedure
end;

exports
Execute,
AllowedComponent,
GetParamCount,
GetParamName,
GetParamUnit,
SetParamValue,
GetParamDefaultValue,
GetErrorMsg;

begin
Frequency:=50;

ListOfComponentsSimu:=TList.Create;
NetNodesList:=TList.Create;

SaveExit := ExitProc;  // Save output procededure
ExitProc := @LibExit;  // Install LibExit exit procedure
end.
