library DC;

{ Network Drawing program }

{ By Phil Marty <phmarty@yahoo.fr> }

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

{
This is then DC engine plugin to use whith Fendi
It is the first very simple DC simulation engine as exemple
}

uses
  SysUtils,
  Classes;

const
   MaxLines=1000;

   {$I '../CompoConst.lst'}

type
  TLigInteger=array[0..999999] of Integer;
  PLigInteger=^TLigInteger;

  TLigDouble=array[0..999999] of Double;
  PLigDouble=^TLigDouble;
  TMatDouble=array[0..999999] of PLigDouble;
  PMatDouble=^TMatDouble;

  // Type used to communication between Fendi and the engines
  // It is quite a text file
  TComm=record
    Count:Integer; // 32 bits
    Lines:array[0..MaxLines-1] of Pchar;
    end;
  PComm=^TComm; // Pointer

  // Common father
  TObjetSimu=class
  end;

  // The net nodes
  TNodeSimu=class(TObjetSimu)
    Number:Integer;
    Father:TObjetSimu;
    procedure SetNumber(_Number:Integer);
    function  GetNumber:Integer;
    procedure SetFather(_Father:TObjetSimu);
    function  GetFather:TObjetSimu;
  end;

  // The components
  TComponentSimu=class(TObjetSimu)
    Name:string;
    ObjetType:Integer;
    NodesList:TList;
    NbParameter:Integer;
    ParameterName:array[0..10] of string; // Will be dynamic later
    Parameter:array[0..10] of Double;     // Will be dynamic later

    constructor Create;
    destructor Destroy;

    procedure SetName(NewName:string);
    function  GetName:string;
    procedure SetType(NewType:Integer);
    function  GetType:Integer;

    procedure SetNode(NodeSimu:TNodeSimu;NodeIndex:Integer);
    function  GetNode(i:Integer):TNodeSimu;

    procedure SetParametersCount(Count:Integer);
    function  GetParametersCount:Integer;
    procedure SetParameterName(i:Integer;_ParameterName:string);
    function  GetParameterName(i:Integer):string;
    procedure SetParameterValue(i:Integer;Value:Double);
    function  GetParameterValue(i:Integer):Double;
  end;

var
   SaveExit: Pointer;

   NetNodesList:TList;
   ListOfComponentsSimu:TList;
   NbActiveComponents:Integer;
   ListActive:TList;
   NbNodes:Integer;
   ErrorMsg:string;

   MatDouble1:PMatDouble;
   MatDouble2:PMatDouble;

{$R *.RES}

// Allocation matrix
procedure GetmemMat(var MatDouble:PMatDouble;
                    Sx,Sy:Integer);
var
   j:Integer;
begin
Getmem(MatDouble,Sy*4);
for j:=0 to Sy-1 do
   begin
   Getmem(MatDouble^[j],Sx*8);
   FillChar(MatDouble^[j]^,Sx*8,0);
   end;
end;

// Free matrix
procedure FreememMat(var MatDouble:PMatDouble;
                     Sx,Sy:Integer);
var
   j:Integer;
begin
for j:=0 to Sy-1 do Freemem(MatDouble^[j],Sx*8);
Freemem(MatDouble,Sy*4);
end;

function GaussJordan(var a:PMatDouble; n:Integer; var b:PMatDouble; m:Integer):Boolean;
var
big,dum,pivinv:Double;
i,icol,irow,j,k,l,ll:Integer;
indxc,indxr,ipiv:PLigInteger;
begin
Getmem(indxc,n*8);
Getmem(indxr,n*8);
Getmem(ipiv,n*8);
Result:=True;
for j:=0 to n-1 do ipiv^[j]:=0;
for i:=0 to n-1 do
   begin
   big:=0;
   for j:=0 to n-1 do
      if ipiv[j]<>1 then
         for k:=0 to n-1 do
            if ipiv^[k]=0 then
               if abs(a^[j]^[k])>=big then
                  begin
                  big:=abs(a^[j]^[k]);
                  irow:=j;
                  icol:=k;
                  end
               else if ipiv^[k]>1 then begin Result:=False; Exit; end;
   ipiv^[icol]:=ipiv^[icol]+1;
   if irow<>icol then
      begin
      for l:=0 to n-1 do
         begin
         dum:=a^[irow]^[l];
         a^[irow]^[l]:=a^[icol]^[l];
         a^[icol]^[l]:=dum;
         end;
      for l:=0 to m-1 do
         begin
         dum:=b^[irow]^[l];
         b^[irow]^[l]:=b^[icol]^[l];
         b^[icol]^[l]:=dum;
         end;
      end;
   indxr^[i]:=irow;
   indxc^[i]:=icol;
   if abs(a[icol,icol])<1e-18 then begin Result:=False; Exit; end;
   pivinv:=1/a^[icol]^[icol];
   a^[icol]^[icol]:=1;
   for l:=0 to n-1 do a^[icol]^[l]:=a^[icol]^[l]*pivinv;
   for l:=0 to m-1 do b^[icol]^[l]:=b^[icol]^[l]*pivinv;
   for ll:=0 to n-1 do
      if ll<>icol then
         begin
         dum:=a^[ll]^[icol];
         a^[ll]^[icol]:=0;
         for l:=0 to n-1 do
            a^[ll]^[l]:=a^[ll]^[l]-a^[icol]^[l]*dum;
         for l:=0 to m-1 do
            b^[ll]^[l]:=b^[ll]^[l]-b^[icol]^[l]*dum;
         end;
   end;
for l:=n-1 downto 0 do
   if indxr^[l]<>indxc^[l] then
      for k:=0 to n-1 do
         begin
         dum:=a^[k]^[indxr^[l]];
         a^[k]^[indxr^[l]]:=a^[k]^[indxc^[l]];
         a^[k]^[indxc^[l]]:=dum;
         end;
Result:=True;
end;

//******************************************************************************
//***************************    TNodeSimu    **********************************
//******************************************************************************

procedure TNodeSimu.SetNumber(_Number:Integer);
begin
Number:=_Number;
end;

function TNodeSimu.GetNumber:Integer;
begin
Result:=Number;
end;

procedure TNodeSimu.SetFather(_Father:TObjetSimu);
begin
Father:=_Father;
end;

function TNodeSimu.GetFather:TObjetSimu;
begin
Result:=Father;
end;

//******************************************************************************
//***********************       TComponentSimu     *****************************
//******************************************************************************

constructor TComponentSimu.Create;
begin
NodesList:=TList.Create;
end;

destructor TComponentSimu.Destroy;
begin
NodesList.Free;
end;

function TComponentSimu.GetName:string;
begin
Result:=Name;
end;

procedure TComponentSimu.SetName(NewName:string);
begin
Name:=NewName;
end;

procedure TComponentSimu.SetType(NewType:Integer);
begin
ObjetType:=NewType;
end;

function TComponentSimu.GetType:Integer;
begin
Result:=ObjetType;
end;

procedure TComponentSimu.SetNode(NodeSimu:TNodeSimu;NodeIndex:Integer);
begin
while NodeIndex>NodesList.Count-1 do NodesList.Add(nil);
NodesList[NodeIndex]:=NodeSimu;
end;

function TComponentSimu.GetNode(i:Integer):TNodeSimu;
begin
Result:=NodesList[i];
end;

procedure TComponentSimu.SetParametersCount(Count:Integer);
begin
NbParameter:=Count;
end;

function TComponentSimu.GetParametersCount:Integer;
begin
Result:=NbParameter;
end;

procedure TComponentSimu.SetParameterName(i:Integer;_ParameterName:string);
begin
ParameterName[i]:=_ParameterName;
end;

function TComponentSimu.GetParameterName(i:Integer):string;
begin
Result:=ParameterName[i];
end;

procedure TComponentSimu.SetParameterValue(i:Integer;Value:Double);
begin
Parameter[i]:=Value;
end;

function TComponentSimu.GetParameterValue(i:Integer):Double;
begin
Result:=Parameter[i];
end;

//******************************************************************************
//**************************  DC Simulation Engine   ***************************
//******************************************************************************

// Load data in Comm
//function Load(var Comm:PComm):Boolean;
function Load:Boolean;
var
   T:TextFile;
   i,j,k:Integer;
   PosX,PosY:integer;
   Path,Line:string;
   NetNode:TList;
   Typ:Integer;
   CompoName:string;
   ComponentSimu:TComponentSimu;
   NodeList:TList;
   NodeIndex:Integer;
   Father:TComponentSimu;
   NodeSimu:TNodeSimu;
begin
AssignFile(T,'datas.tmp');
Reset(T);
try

//k:=0;
Result:=True;

//<?xml version='1.0' encoding='UTF16' standalone='yes'?>
//Inc(k);
Readln(T,Line);

//<component>
//Line:=Comm.Lines[k];
//Inc(k);
Readln(T,Line);
if Line<>'<component>' then
   begin
   ErrorMsg:='Error : No component in the file';
   Result:=False;
   Exit;
   end;

// Load components
while Line='<component>' do
   begin
   //<type>Objet.GetType</type>
   //Line:=Comm.Lines[k];
   //Inc(k);
   Readln(T,Line);
   Typ:=StrToInt(Copy(Line,7,Length(Line)-13));
   //<name>Objet.GetName</name>
   //Line:=Comm.Lines[k];
   //Inc(k);
   Readln(T,Line);
   CompoName:=Copy(Line,7,Length(Line)-13);

   ComponentSimu:=TComponentSimu.Create;
   ComponentSimu.SetType(Typ);
   ComponentSimu.SetName(CompoName);
   ListOfComponentsSimu.Add(ComponentSimu);

   //<parameternumber>Objet.GetParametersCount</parameternumber>
   //Line:=Comm.Lines[k];
   //Inc(k);
   Readln(T,Line);
   Line:=Copy(Line,18,Length(Line)-35);
   ComponentSimu.SetParametersCount(StrToInt(Line));
   for i:=0 to ComponentSimu.GetParametersCount-1 do
      begin
      //<parametername>Objet.GetParameterName()</parametername>
      //Line:=Comm.Lines[k];
      //Inc(k);
      Readln(T,Line);
      Line:=Copy(Line,16,Length(Line)-31);
      ComponentSimu.SetParameterName(i,Line);
      //<parametervalue>Objet.GetParameter()</parametervalue>
      //Line:=Comm.Lines[k];
      //Inc(k);
      Readln(T,Line);
      Line:=Copy(Line,17,Length(Line)-33);
      ComponentSimu.SetParameterValue(i,StrToFloat(Line));
      end;

   //Inc(k);
   Readln(T,Line);
   //Line:=Comm.Lines[k];
   //Inc(k);
   Readln(T,Line);
   end;

// Load nodes and links to components
//<netnode>
if Line<>'<netnode>' then
   begin
   ErrorMsg:='Error : No netnode in the file';
   Result:=False;
   Exit;
   end;

while Line='<netnode>' do
   begin
   //<node>
   //Line:=Comm.Lines[k];
   //Inc(k);
   Readln(T,Line);
   if Line<>'<node>' then
      begin
      ErrorMsg:='Error : No node in netnode';
      Result:=False;
      Exit;
      end;

   NodeList:=TList.Create;
   NetNodesList.Add(NodeList);
   while Line='<node>' do
      begin
      //<component>Component.GetName</component>
      //Line:=Comm.Lines[k];
      //Inc(k);
      Readln(T,Line);
      CompoName:=Copy(Line,12,Length(Line)-23);
      //<index>Node.GetIndex</index>
      //Line:=Comm.Lines[k];
      //Inc(k);
      Readln(T,Line);
      NodeIndex:=StrToInt(Copy(Line,8,Length(Line)-15));

      // Get ComponentSimu with name CompoName
      for i:=0 to ListOfComponentsSimu.Count-1 do
         begin
         ComponentSimu:=ListOfComponentsSimu[i];
         if ComponentSimu.GetName=CompoName then Father:=ComponentSimu;
         end;

      NodeSimu:=TNodeSimu.Create;
      NodeSimu.SetNumber(NodeIndex);
      NodeSimu.SetFather(Father);
      Father.SetNode(NodeSimu,NodeIndex);
      NodeList.Add(NodeSimu);

      //Inc(k);
      Readln(T,Line);
      //Line:=Comm.Lines[k];
      //Inc(k);
      Readln(T,Line);
      end;
   //if k<Comm.Count then Line:=Comm.Lines[k];
   //Inc(k);
   Readln(T,Line);
   end;

finally
CloseFile(T);
end;
end;

// Save results in Comm
//function SaveResults(var Comm:PComm):Boolean;
function SaveResults:Boolean;
var
   i:Integer;
   ComponentSimu:TComponentSimu;
   NetNode:TList;
   T:TextFile;

   procedure AddLine(StrVal:string);
   begin
//   Getmem(Comm.Lines[Comm.Count],Length(StrVal));
//   StrCopy(Comm.Lines[Comm.Count],PChar(StrVal));
//   Inc(Comm.Count);
   Writeln(T,StrVal);   
   end;
   
begin
AssignFile(T,'results.tmp');
Rewrite(T);
try


//New(Comm);
//Comm.Count:=0;

AddLine('<?xml version=''1.0'' encoding=''UTF16'' standalone=''yes''?>');

// Save components infos
for i:=0 to NbActiveComponents-1 do
   begin
   ComponentSimu:=ListActive[i];
   AddLine('<component>');
   AddLine('<type>'+IntToStr(ComponentSimu.GetType)+'</type>');
   AddLine('<name>'+ComponentSimu.GetName+'</name>');
   AddLine('<infosnumber>1</infosnumber>');
   AddLine('<infoname>Current</infoname>');
   AddLine('<infotype>0</infotype>'); // Real
   AddLine('<infounit>A</infounit>');
   AddLine('<infovalue>'+FloatToStr(MatDouble2^[i*2]^[0])+'</infovalue>');
   AddLine('</component>');
   end;

// Save nodes infos
for i:=0 to NbNodes-1 do
   begin
   NetNode:=NetNodesList[i];
   // Only first node of the net node / The others are the same
   AddLine('<netnode>');
   AddLine('<infosnumber>1</infosnumber>');
   AddLine('<infoname>Voltage</infoname>');
   AddLine('<infotype>0</infotype>'); // Real
   AddLine('<infounit>V</infounit>');
   AddLine('<infovalue>'+FloatToStr(MatDouble2^[NbActiveComponents*2+i]^[0])+'</infovalue>');
   AddLine('</netnode>');
   end;

// Todo : Lines ?

finally
CloseFile(T);
end;
end;

//function Execute(var Comm:PComm):Boolean;
function Execute:Boolean;
var
   i,j,k,l:Integer;
   ComponentSimu:TComponentSimu;
   Dim:Integer;
   Node1,Node2:TNodeSimu;
   NetNode:TList;
   NodeNumber1,NodeNumber2,CompoNumber,NeutralNode,CompoType:Integer;
   T:TextFile;
begin
// Free old data
for i:=0 to ListOfComponentsSimu.Count-1 do
   begin
   ComponentSimu:=ListOfComponentsSimu[i];
   ComponentSimu.Free;
   end;
ListOfComponentsSimu.Clear;

for j:=0 to NetNodesList.Count-1 do
   begin
   NetNode:=NetNodesList[j];
   NetNode.Free;
   end;
NetNodesList.Clear;

// Load new data
//if not Load(Comm) then begin Result:=False; Exit; end;
if not Load then begin Result:=False; Exit; end;

// Free Comm
//for i:=0 to Comm.Count-1 do
//   StrDispose(Comm.Lines[i]);
//Comm.Count:=0;

// More than one neutral forbidden
j:=0;
for i:=0 to ListOfComponentsSimu.Count-1 do
   begin
   ComponentSimu:=ListOfComponentsSimu[i] ;
   if ComponentSimu.GetType=ncNeutral then Inc(j);
   end;

if j>1 then
   begin
   ErrorMsg:='Error : Only one neutral allowed';
   Exit;
   end;

// List of active components
ListActive:=TList.Create;

// Find active components
j:=0;
for i:=0 to ListOfComponentsSimu.Count-1 do
   begin
   ComponentSimu:=ListOfComponentsSimu[i];
   if ComponentSimu.GetType<>ncNeutral then
      begin
      Inc(j);
      ListActive.Add(ComponentSimu);
      end;
   end;

// Find matrix dimension
NbNodes:=NetNodesList.Count;
NbActiveComponents:=j;
//Dim:=NbNodes+NbActiveComponents;
Dim:=NbNodes+NbActiveComponents*2; // 2 currents by component

// Create network matrix in the heap
GetmemMat(MatDouble1,Dim,Dim);
GetmemMat(MatDouble2,1,Dim);

// Matrix form :
// i1 i1' i2 i2' i3 i3' ......  v1 v2 v3 ......

//             |----------|
//  ----->-----|          |-----<-----
//       i     |----------|     i'

//*****************************************
//********** Component equations **********
//*****************************************
for i:=0 to NbActiveComponents-1 do
   begin
   ComponentSimu:=TComponentSimu(ListActive[i]);
   // For this firts DC engine, we take the first parameter of components
   // because there is only one parameter in components.
   case ComponentSimu.GetType of
      ncCapacitor,ncResistor,ncSourceDC:
         begin
         // ****** The parameter part of the equation ******
         if ComponentSimu.GetType=ncResistor then
            MatDouble1^[i]^[i*2]:=ComponentSimu.GetParameterValue(0);
         if ComponentSimu.GetType=ncSourceDC then
            MatDouble2^[i]^[0]:=ComponentSimu.GetParameterValue(0);

         // ****** The voltage part of the equation ******
         // Only resistors with 2 nodes components in this engine
         // With others components and more nodes in component we'll have to
         // rewrite this part. The relationships between voltages and current
         // will be depend of the component
         // Find nodes
         Node1:=ComponentSimu.GetNode(0);
         // Find corresponding netnode
         for j:=0 to NetNodesList.Count-1 do
            begin
            NetNode:=NetNodesList[j];
            for k:=0 to NetNode.Count-1 do
               begin
               if Node1=NetNode[k] then
                  begin
                  NodeNumber1:=j;
                  end
               end;
            end;

         // A -1 for this node
         // For a source it is inverted
         if ComponentSimu.GetType=ncSourceDC then
            MatDouble1^[i]^[NbActiveComponents*2+NodeNumber1]:=1
         else
            MatDouble1^[i]^[NbActiveComponents*2+NodeNumber1]:=-1;

         Node2:=ComponentSimu.GetNode(1);
         // Find corresponding netnode
         for j:=0 to NetNodesList.Count-1 do
            begin
            NetNode:=NetNodesList[j];
            for k:=0 to NetNode.Count-1 do
               begin
               if Node2=NetNode[k] then
                  begin
                  NodeNumber1:=j;
                  end
               end;
            end;

         // A 1 for this node
         // For a source it is inverted
         if ComponentSimu.GetType=ncSourceDC then
            MatDouble1^[i]^[NbActiveComponents*2+NodeNumber1]:=-1
         else
            MatDouble1^[i]^[NbActiveComponents*2+NodeNumber1]:=1;
         end;
      end;
   end;

//**************************************************
//********** Components current equations **********
//**************************************************
// i'+i=0

// Here, it is the same for all components
// But it may change later
for i:=0 to NbActiveComponents-1 do
   begin
   MatDouble1^[NbActiveComponents+i]^[i*2]:=1;
   MatDouble1^[NbActiveComponents+i]^[i*2+1]:=1;
   end;

//************************************
//********** Node equations **********
//************************************
for i:=0 to NbNodes-2 do
   begin
   // Find components connected to the node
   NetNode:=NetNodesList[i];
   for j:=0 to NetNode.Count-1 do
      begin
      Node1:=NetNode[j];
      ComponentSimu:=Node1.Father as TComponentSimu;
      if ComponentSimu.GetType<>ncConnection then
         begin
         // Find component number
         l:=0;
         for k:=0 to NbActiveComponents-1 do
            begin
            ComponentSimu:=TComponentSimu(ListActive[k]);
            if ComponentSimu=Node1.Father then
               begin
               CompoNumber:=l;
               CompoType:=ComponentSimu.GetType;
               end;
            inc(l);
            end;

         // Find the side of the component (Only 2 node component in this engine) for the sign
         // because we calculate only one current in this engine
         // The current are equals but opposed in the two nodes
         // For more then 2 node, we'll have to use one current for each component node
         // For a source it is a generator convention
         if Node1.GetNumber=0 then
            begin
            if CompoType=ncSourceDC then
               MatDouble1^[NbActiveComponents*2+i]^[CompoNumber*2]:=-1
            else
               MatDouble1^[NbActiveComponents*2+i]^[CompoNumber*2]:=1;
            end
         else
            begin
            if CompoType=ncSourceDC then
               MatDouble1^[NbActiveComponents*2+i]^[CompoNumber*2]:=-1
            else
               MatDouble1^[NbActiveComponents*2+i]^[CompoNumber*2]:=1;
            end;
         end;
      end;

   end;

//***************************************
//********** Neutral equations **********
//***************************************
// Last special line for the neutral node
// Find neutral node
for i:=0 to NbNodes-1 do
   begin
   // Find components connected to the node
   NetNode:=NetNodesList[i];
   for j:=0 to NetNode.Count-1 do
      begin
      Node1:=NetNode[j];
      // Save neutral node for later
      ComponentSimu:=Node1.Father as TComponentSimu;
      if ComponentSimu.GetType=ncNeutral then NeutralNode:=i;
      end;
   end;
MatDouble1^[Dim-1]^[NbActiveComponents*2+NeutralNode]:=1;

// Save results to a file to check
AssignFile(T,'Matrix.txt');
Rewrite(T);
try

for i:=0 to Dim-1 do
   begin
   for j:=0 to Dim-1 do
      Write(T,Format('%2.4f',[MatDouble1^[i]^[j]])+'   ');
   Write(T,Format('%2.4f',[MatDouble2^[i]^[0]])+'   ');
   Writeln(T,'');
   end;

// Solve network matrix
GaussJordan(MatDouble1,Dim,MatDouble2,1);

Writeln(T,'');
for i:=0 to Dim-1 do
   begin
   for j:=0 to Dim-1 do
      Write(T,Format('%2.4f',[MatDouble1^[i]^[j]])+'   ');
   Write(T,Format('%2.4f',[MatDouble2^[i]^[0]])+'   ');
   Writeln(T,'');
   end;

finally
CloseFile(T);
end;

// Save results
//SaveResults(Comm);
SaveResults;

ListActive.Free;
end;

// List of allowed components for this engine
function AllowedComponent(Number:Integer):Boolean;
begin
Result:=False;
case Number of
   ncNone       :Result:=True;
   ncNode       :Result:=True;
   ncLine       :Result:=True;
   ncConnection :Result:=True;
   ncResistor   :Result:=True;
   ncSourceDC   :Result:=True;
   ncNeutral    :Result:=True;
   ncComment    :Result:=True;
   end;
end;

function GetParamCount:Integer;
begin
Result:=0;
end;

function GetParamName(i:Integer):PChar;
begin
Result:='';
end;

function GetParamUnit(i:Integer):PChar;
begin
Result:='';
end;

procedure SetParamValue(i:Integer;Value:Double);
begin
end;

function GetParamDefaultValue(i:Integer):Double;
begin
Result:=0;
end;

function GetErrorMsg:PChar;
begin
Result:=PChar(ErrorMsg);
end;

procedure LibExit;
begin
// Library exit code
ListOfComponentsSimu.Free;
ExitProc := SaveExit;  // Restore output procedure
end;

exports
Execute,
AllowedComponent,
GetParamCount,
GetParamName,
GetParamUnit,
SetParamValue,
GetParamDefaultValue,
GetErrorMsg;

begin
ListOfComponentsSimu:=TList.Create;
NetNodesList:=TList.Create;

SaveExit := ExitProc;  // Save output procededure
ExitProc := @LibExit;  // Install LibExit exit procedure
end.
